/**
 * @file
 * @brief Provides a logger and macros for convenient access
 * @copyright Copyright (c) 2017 CERN and the Allpix Squared authors.
 * This software is distributed under the terms of the MIT License, copied
 * verbatim in the file "LICENSE.md".
 * In applying this license, CERN does not waive the privileges and immunities
 * granted to it by virtue of its status as an
 * Intergovernmental Organization or submit itself to any jurisdiction.
 */

#ifndef CORRYVRECKAN_LOG_H
#define CORRYVRECKAN_LOG_H

#ifdef WIN32
#define __func__ __FUNCTION__
#endif

#include <cstring>
#include <mutex>
#include <ostream>
#include <sstream>
#include <string>
#include <vector>

namespace corryvreckan {
    /**
     * @brief Logging detail level
     */
    enum class LogLevel {
        FATAL = 0, ///< Fatal problems that terminate the framework (typically exceptions)
        STATUS,    ///< Only critical progress information
        ERROR,     ///< Critical problems that usually lead to fatal errors
        WARNING,   ///< Possible issue that could lead to unexpected results
        INFO,      ///< General information about processes (should not be called in run
                   /// function)
        DEBUG,     ///< Detailed information about physics process
        NONE,      ///< Indicates the log level has not been set (cannot be selected by the
                   /// user)
        TRACE,     ///< Software debugging information about what part is currently
                   /// running
    };
    /**
     * @brief Format of the logger
     */
    enum class LogFormat {
        SHORT = 0, ///< Only include a single character for the log level, the section
                   /// header and the message
        DEFAULT,   ///< Also include the time and a full logging level description
        LONG       ///< All of the above and also information about the file and line where
                   /// the message was defined
    };

    /**
     * @brief Logger of the framework to inform the user of process
     *
     * Should almost never be instantiated directly. The \ref LOG macro should be
     * used instead to pass all the information.
     * This leads to a cleaner interface for sending log messages.
     */
    // TODO [DOC] This just be renamed to Log?
    class DefaultLogger {
    public:
        /**
         * @brief Construct a logger
         */
        DefaultLogger();
        /**
         * @brief Write the output to the streams and destruct the logger
         */
        ~DefaultLogger();

        ///@{
        /**
         * @brief Disable copying
         */
        DefaultLogger(const DefaultLogger&) = delete;
        DefaultLogger& operator=(const DefaultLogger&) = delete;
        ///@}

        ///@{
        /**
         * @brief Use default move behaviour
         */
        DefaultLogger(DefaultLogger&&) noexcept(false) = default;
        DefaultLogger& operator=(DefaultLogger&&) noexcept(false) = default;
        ///@}

        /**
         * @brief Gives a stream to write to using the C++ stream syntax
         * @param level Logging level
         * @param file The file name of the file containing the log message
         * @param function The function containing the log message
         * @param line The line number of the log message
         * @return A C++ stream to write to
         */
        std::ostringstream& getStream(LogLevel level = LogLevel::INFO,
                                      const std::string& file = "",
                                      const std::string& function = "",
                                      uint32_t line = 0);

        /**
         * @brief Gives a process stream which updates the same line as long as it is
         * the same
         * @param identifier Name to indicate the line, used to distinguish when to
         * update or write new line
         * @param level Logging level
         * @param file The file name of the file containing the log message
         * @param function The function containing the log message
         * @param line The line number of the log message
         * @return A C++ stream to write to
         */
        std::ostringstream& getProcessStream(std::string identifier,
                                             LogLevel level = LogLevel::INFO,
                                             const std::string& file = "",
                                             const std::string& function = "",
                                             uint32_t line = 0);

        /**
         * @brief Finish the logging ensuring proper termination of all streams
         */
        static void finish();

        /**
         * @brief Get the reporting level for logging
         * @return The current log level
         */
        static LogLevel getReportingLevel();
        /**
         * @brief Set a new reporting level to use for logging
         * @param level The new log level
         */
        static void setReportingLevel(LogLevel level);

        /**
         * @brief Convert a string to a LogLevel
         * @param level Name of the level
         * @return Log level corresponding to the name
         */
        static LogLevel getLevelFromString(const std::string& level);
        /**
         * @brief Convert a LogLevel to a string
         * @param level Log level
         * @return Name corresponding to the log level
         */
        static std::string getStringFromLevel(LogLevel level);

        /**
         * @brief Get the logging format
         * @return Current log format
         */
        static LogFormat getFormat();
        /**
         * @brief Set a new logging format
         * @param format New log log format
         */
        static void setFormat(LogFormat format);

        /**
         * @brief Convert a string to a LogFormat
         * @param format Name of the format
         * @return Log format corresponding to the name
         */
        static LogFormat getFormatFromString(const std::string& format);
        /**
         * @brief Convert a LogFormat to a string
         * @param format Log format
         * @return Name corresponding to the log format
         */
        static std::string getStringFromFormat(LogFormat format);

        /**
         * @brief Add a stream to write the log message
         * @param stream Stream to write to
         */
        static void addStream(std::ostream& stream);
        /**
         * @brief Clear and delete all streams that the logger writes to
         */
        static void clearStreams();
        /**
         * @brief Return all the streams the logger writes to
         * @return List of all the streams used
         */
        static const std::vector<std::ostream*>& getStreams();

        /**
         * @brief Set the section header to use from now on
         * @param header Header to use
         */
        static void setSection(std::string header);
        /**
         * @brief Get the current section header
         * @return Header used
         */
        static std::string getSection();

    private:
        /**
         * @brief The number of exceptions that are uncaught
         * @param cons If true: always return zero if amount of exceptions cannot be
         * properly determined
         * @return Number of uncaught exceptions
         */
        int get_uncaught_exceptions(bool cons);
        /**
         * @brief Get the current date as a printable string
         * @return Current date as a string
         */
        std::string get_current_date();

        /**
         * @brief Return if a stream is likely a terminal screen (supporting colors
         * etc.)
         * @return True if the stream is terminal, false otherwise
         */
        static bool is_terminal(std::ostream& stream);

        // Output stream
        std::ostringstream os;

        // Number of exceptions to prevent abort
        int exception_count_{};
        // Saved value of the length of the header indent
        unsigned int indent_count_{};

        // Internal methods to store static values
        static std::string& get_section();
        static LogLevel& get_reporting_level();
        static LogFormat& get_format();
        static std::vector<std::ostream*>& get_streams();

        // Name of the process to log or empty if a normal log message
        std::string identifier_{};
        static std::string last_message_;
        static std::string last_identifier_;

        static std::mutex write_mutex_;
    };

    using Log = DefaultLogger;

/**
 *  @brief Base name of the file without the directory
 */
#define __FILE_NAME__ (strrchr(__FILE__, '/') ? strrchr(__FILE__, '/') + 1 : __FILE__)

/**
 * @brief Execute a block only if the reporting level is high enough
 * @param level The minimum log level
 */
#define IFLOG(level)                                                                                                        \
    if(corryvreckan::LogLevel::level <= corryvreckan::Log::getReportingLevel() && !corryvreckan::Log::getStreams().empty())

/**
 * @brief Create a logging stream if the reporting level is high enough
 * @param level The log level of the stream
 */
#define LOG(level)                                                                                                          \
    if(corryvreckan::LogLevel::level <= corryvreckan::Log::getReportingLevel() && !corryvreckan::Log::getStreams().empty()) \
    corryvreckan::Log().getStream(                                                                                          \
        corryvreckan::LogLevel::level, __FILE_NAME__, std::string(static_cast<const char*>(__func__)), __LINE__)

/**
 * @brief Create a logging stream that overwrites the line if the previous
 * message has the same identifier
 * @param level The log level of the stream
 * @param identifier Identifier for this stream to determine overwrites
 */
#define LOG_PROGRESS(level, identifier)                                                                                     \
    if(corryvreckan::LogLevel::level <= corryvreckan::Log::getReportingLevel() && !corryvreckan::Log::getStreams().empty()) \
    corryvreckan::Log().getProcessStream(identifier,                                                                        \
                                         corryvreckan::LogLevel::level,                                                     \
                                         __FILE_NAME__,                                                                     \
                                         std::string(static_cast<const char*>(__func__)),                                   \
                                         __LINE__)

    /**
     * @brief Suppress an stream from writing any output
     * @param stream The stream to suppress
     */
    // suppress a (logging) stream
    // TODO [doc] rewrite as a lowercase function in a namespace?
    inline void SUPPRESS_STREAM(std::ostream& stream) { stream.setstate(std::ios::failbit); }

    /**
     * @brief Release an suppressed stream so it can write again
     * @param stream The stream to release
     */
    // TODO [doc] rewrite as a lowercase function in a namespace?
    inline void RELEASE_STREAM(std::ostream& stream) { stream.clear(); }
} // namespace corryvreckan

#endif /* CORRYVRECKAN_LOG_H */
